=== modified file 'drizzled/db.cc'
--- drizzled/db.cc	2010-02-25 07:52:23 +0000
+++ drizzled/db.cc	2010-02-25 20:20:08 +0000
@@ -133,7 +133,7 @@
   }
   else // Created !
   {
-    replication_services.rawStatement(session, session->query);
+    replication_services.createSchema(session, schema_message);
     session->my_ok(1);
   }
 

=== modified file 'drizzled/message/statement_transform.cc'
--- drizzled/message/statement_transform.cc	2010-02-08 01:10:03 +0000
+++ drizzled/message/statement_transform.cc	2010-02-25 20:20:08 +0000
@@ -2,10 +2,11 @@
  *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
  *
  *  Copyright (C) 2009 Sun Microsystems
+ *  Copyright (c) 2010 Jay Pipes
  *
  *  Authors:
  *
- *    Jay Pipes <joinfu@sun.com>
+ *    Jay Pipes <jaypipes@gmail.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -194,6 +195,16 @@
       sql_strings.push_back(destination);
     }
     break;
+  case message::Statement::CREATE_SCHEMA:
+    {
+      assert(source.has_create_schema_statement());
+      string destination;
+      error= message::transformCreateSchemaStatementToSql(source.create_schema_statement(),
+                                                          &destination,
+                                                          sql_variant);
+      sql_strings.push_back(destination);
+    }
+    break;
   case message::Statement::SET_VARIABLE:
     {
       assert(source.has_set_variable_statement());
@@ -487,8 +498,6 @@
     if (should_quote_field_value)
       destination->push_back('\'');
   }
-  if (num_key_fields > 1)
-    destination->push_back(')');
 
   return error;
 }
@@ -646,6 +655,31 @@
 }
 
 enum message::TransformSqlError
+message::transformCreateSchemaStatementToSql(const message::CreateSchemaStatement &statement,
+                                             std::string *destination,
+                                             enum message::TransformSqlVariant sql_variant)
+{
+  char quoted_identifier= '`';
+  if (sql_variant == ANSI)
+    quoted_identifier= '"';
+
+  const message::Schema &schema= statement.schema();
+
+  destination->append("CREATE SCHEMA ", 14);
+  destination->push_back(quoted_identifier);
+  destination->append(schema.name());
+  destination->push_back(quoted_identifier);
+
+  if (schema.has_collation())
+  {
+    destination->append(" COLLATE ", 9);
+    destination->append(schema.collation());
+  }
+
+  return NONE;
+}
+
+enum message::TransformSqlError
 message::transformTruncateTableStatementToSql(const message::TruncateTableStatement &statement,
                                               std::string *destination,
                                               enum message::TransformSqlVariant sql_variant)

=== modified file 'drizzled/message/statement_transform.h'
--- drizzled/message/statement_transform.h	2010-02-08 01:10:03 +0000
+++ drizzled/message/statement_transform.h	2010-02-25 20:20:08 +0000
@@ -53,6 +53,7 @@
 class DeleteData;
 class DeleteRecord;
 class TruncateTableStatement;
+class CreateSchemaStatement;
 class SetVariableStatement;
 
 /** A Variation of SQL to be output during transformation */
@@ -283,6 +284,25 @@
                                      enum TransformSqlVariant sql_variant= DRIZZLE);
 
 /**
+ * This function looks at a supplied CreateSchemaStatement
+ * and constructs a correctly-formatted SQL
+ * statement to the supplied destination string.
+ *
+ * @param CreateSchemaStatement message to transform
+ * @param Destination string to append SQL to
+ * @param Variation of SQL to generate
+ *
+ * @retval
+ *  NONE if successful transformation
+ * @retval
+ *  Error code (see enum TransformSqlError definition) if failure
+ */
+enum TransformSqlError
+transformCreateSchemaStatementToSql(const CreateSchemaStatement &statement,
+                                    std::string *destination,
+                                    enum TransformSqlVariant sql_variant= DRIZZLE);
+
+/**
  * This function looks at a supplied SetVariableStatement
  * and constructs a correctly-formatted SQL
  * statement to the supplied destination string.

=== modified file 'drizzled/replication_services.cc'
--- drizzled/replication_services.cc	2010-02-11 03:49:10 +0000
+++ drizzled/replication_services.cc	2010-02-25 20:20:08 +0000
@@ -508,7 +508,7 @@
      * We add the "key field metadata" -- i.e. the fields which is
      * the primary key for the table.
      */
-    if (in_table->s->primary_key == current_field->field_index)
+    if (in_table->s->fieldInPrimaryKey(current_field))
     {
       field_metadata= header->add_key_field_metadata();
       field_metadata->set_name(current_field->field_name);
@@ -602,7 +602,7 @@
      * primary key field value.  Replication only supports tables
      * with a primary key.
      */
-    if (in_table->s->primary_key == current_field->field_index)
+    if (in_table->s->fieldInPrimaryKey(current_field))
     {
       /**
        * To say the below is ugly is an understatement. But it works.
@@ -682,7 +682,7 @@
      * primary key field value.  Replication only supports tables
      * with a primary key.
      */
-    if (in_table->s->primary_key == current_field->field_index)
+    if (in_table->s->fieldInPrimaryKey(current_field))
     {
       field_metadata= header->add_key_field_metadata();
       field_metadata->set_name(current_field->field_name);
@@ -715,7 +715,7 @@
      * primary key field value.  Replication only supports tables
      * with a primary key.
      */
-    if (in_table->s->primary_key == current_field->field_index)
+    if (in_table->s->fieldInPrimaryKey(current_field))
     {
       string_value= current_field->val_str(string_value);
       record->add_key_value(string_value->c_ptr(), string_value->length());
@@ -727,6 +727,35 @@
   }
 }
 
+void ReplicationServices::createSchema(Session *in_session,
+                                       const message::Schema &schema)
+{
+  if (! is_active)
+    return;
+  
+  message::Transaction *transaction= getActiveTransaction(in_session);
+  message::Statement *statement= transaction->add_statement();
+
+  initStatement(*statement, message::Statement::CREATE_SCHEMA, in_session);
+
+  /* 
+   * Construct the specialized CreateSchemaStatement message and attach
+   * it to the generic Statement message
+   */
+  message::CreateSchemaStatement *create_schema_statement= statement->mutable_create_schema_statement();
+  message::Schema *new_schema_message= create_schema_statement->mutable_schema();
+  *new_schema_message= schema;
+
+  finalizeStatement(*statement, in_session);
+
+  finalizeTransaction(*transaction, in_session);
+  
+  push(*transaction);
+
+  cleanupTransaction(transaction, in_session);
+
+}
+
 void ReplicationServices::truncateTable(Session *in_session, Table *in_table)
 {
   if (! is_active)

=== modified file 'drizzled/replication_services.h'
--- drizzled/replication_services.h	2010-02-11 03:29:52 +0000
+++ drizzled/replication_services.h	2010-02-25 20:20:08 +0000
@@ -2,10 +2,11 @@
  *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
  *
  *  Copyright (C) 2008-2009 Sun Microsystems
+ *  Copyright (c) 2009-2010 Jay Pipes <jaypipes@gmail.com>
  *
  *  Authors:
  *
- *    Jay Pipes <joinfu@sun.com>
+ *    Jay Pipes <jaypipes@gmail.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,7 +31,6 @@
 
 #include <vector>
 
-
 namespace drizzled
 {
 
@@ -323,7 +323,16 @@
    */
   void deleteRecord(Session *in_session, Table *in_table);
   /**
-   * Creates a TruncateTable Statement GPB message and add it
+   * Creates a CreateSchema Statement GPB message and adds it
+   * to the Session's active Transaction GPB message for pushing
+   * out to the replicator streams.
+   *
+   * @param[in] Pointer to the Session which issued the statement
+   * @param[in] message::Schema message describing new schema
+   */
+  void createSchema(Session *in_session, const message::Schema &schema);
+  /**
+   * Creates a TruncateTable Statement GPB message and adds it
    * to the Session's active Transaction GPB message for pushing
    * out to the replicator streams.
    *

=== modified file 'drizzled/table_share.cc'
--- drizzled/table_share.cc	2010-02-04 08:14:46 +0000
+++ drizzled/table_share.cc	2010-02-25 20:20:08 +0000
@@ -273,4 +273,31 @@
   }
 }
 
+/**
+ * @todo
+ *
+ * Precache this stuff....
+ */
+bool TableShare::fieldInPrimaryKey(Field *in_field) const
+{
+  assert(table_proto != NULL);
+  
+  size_t num_indexes= table_proto->indexes_size();
+
+  for (size_t x= 0; x < num_indexes; ++x)
+  {
+    const message::Table::Index &index= table_proto->indexes(x);
+    if (index.is_primary())
+    {
+      size_t num_parts= index.index_part_size();
+      for (size_t y= 0; y < num_parts; ++y)
+      {
+        if (index.index_part(y).fieldnr() == in_field->field_index)
+          return true;
+      }
+    }
+  }
+  return false;
+}
+
 } /* namespace drizzled */

=== modified file 'drizzled/table_share.h'
--- drizzled/table_share.h	2010-02-23 01:14:53 +0000
+++ drizzled/table_share.h	2010-02-25 20:20:08 +0000
@@ -260,6 +260,12 @@
     max_rows= arg;
   }
 
+  /**
+   * Returns true if the supplied Field object
+   * is part of the table's primary key.
+   */
+  bool fieldInPrimaryKey(Field *field) const;
+
   plugin::StorageEngine *storage_engine;			/* storage engine plugin */
   inline plugin::StorageEngine *db_type() const	/* table_type for handler */
   {

=== modified file 'plugin/logging_gearman/logging_gearman.cc'
--- plugin/logging_gearman/logging_gearman.cc	2010-02-18 04:43:04 +0000
+++ plugin/logging_gearman/logging_gearman.cc	2010-02-25 20:20:08 +0000
@@ -257,7 +257,7 @@
                // dont need to quote the db name, always CSV safe
                (int)session->db.length(), dbs,
                // do need to quote the query
-               quotify((unsigned char *)session->getQueryString(),
+               quotify(reinterpret_cast<const unsigned char *>(session->getQueryString().c_str()),
                        session->getQueryLength(), qs, sizeof(qs)),
                // command_name is defined in drizzled/sql_parse.cc
                // dont need to quote the command name, always CSV safe

=== added file 'plugin/transaction_log/tests/r/create_schema.result'
--- plugin/transaction_log/tests/r/create_schema.result	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/r/create_schema.result	2010-02-25 20:20:08 +0000
@@ -0,0 +1,13 @@
+DROP SCHEMA IF EXISTS my_new_schema;
+CREATE SCHEMA my_new_warnings;
+DROP SCHEMA my_new_warnings;
+START TRANSACTION;
+DROP SCHEMA IF EXISTS my_new_schema;
+COMMIT;
+START TRANSACTION;
+CREATE SCHEMA `my_new_warnings` COLLATE utf8_general_ci;
+COMMIT;
+START TRANSACTION;
+DROP SCHEMA my_new_warnings;
+COMMIT;
+SET GLOBAL transaction_log_truncate_debug= true;

=== modified file 'plugin/transaction_log/tests/r/database.result'
--- plugin/transaction_log/tests/r/database.result	2009-12-23 14:24:22 +0000
+++ plugin/transaction_log/tests/r/database.result	2010-02-25 20:20:08 +0000
@@ -1,7 +1,7 @@
 CREATE DATABASE test1;
 DROP DATABASE test1;
 START TRANSACTION;
-CREATE DATABASE test1;
+CREATE SCHEMA `test1` COLLATE utf8_general_ci;
 COMMIT;
 START TRANSACTION;
 DROP DATABASE test1;

=== modified file 'plugin/transaction_log/tests/r/filtered_replicator.result'
--- plugin/transaction_log/tests/r/filtered_replicator.result	2009-12-23 14:24:22 +0000
+++ plugin/transaction_log/tests/r/filtered_replicator.result	2010-02-25 20:20:08 +0000
@@ -179,7 +179,7 @@
 DROP TABLE `t1`;
 COMMIT;
 START TRANSACTION;
-CREATE DATABASE test1;
+CREATE SCHEMA `test1` COLLATE utf8_general_ci;
 COMMIT;
 START TRANSACTION;
 DROP DATABASE test1;

=== added file 'plugin/transaction_log/tests/r/multi_column_primary_key.result'
--- plugin/transaction_log/tests/r/multi_column_primary_key.result	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/r/multi_column_primary_key.result	2010-02-25 20:20:08 +0000
@@ -0,0 +1,55 @@
+DROP TABLE IF EXISTS t1;
+CREATE TABLE t1 (
+k1 INT NOT NULL
+, k2 INT NOT NULL
+, padding VARCHAR(200) NOT NULL
+, PRIMARY KEY (k1, k2)
+);
+INSERT INTO t1 VALUES (1, 1, "I love testing.");
+INSERT INTO t1 VALUES (2, 2, "I hate testing.");
+INSERT INTO t1 VALUES (2, 3, "I hate and love testing.");
+INSERT INTO t1 VALUES (3, 3, "I adore testing.");
+UPDATE t1 SET padding= "XXX" WHERE k1= 1 AND k2= 1;
+UPDATE t1 SET padding= "YYY" WHERE k1= 2;
+UPDATE t1 SET padding= "ZZZ" WHERE k2= 3;
+UPDATE t1 SET padding= "AAA";
+DROP TABLE t1;
+START TRANSACTION;
+DROP TABLE IF EXISTS `t1`;
+COMMIT;
+START TRANSACTION;
+CREATE TABLE t1 ( k1 INT NOT NULL , k2 INT NOT NULL , padding VARCHAR(200) NOT NULL , PRIMARY KEY (k1, k2) );
+COMMIT;
+START TRANSACTION;
+INSERT INTO `test`.`t1` (`k1`,`k2`,`padding`) VALUES (1,1,'I love testing.');
+COMMIT;
+START TRANSACTION;
+INSERT INTO `test`.`t1` (`k1`,`k2`,`padding`) VALUES (2,2,'I hate testing.');
+COMMIT;
+START TRANSACTION;
+INSERT INTO `test`.`t1` (`k1`,`k2`,`padding`) VALUES (2,3,'I hate and love testing.');
+COMMIT;
+START TRANSACTION;
+INSERT INTO `test`.`t1` (`k1`,`k2`,`padding`) VALUES (3,3,'I adore testing.');
+COMMIT;
+START TRANSACTION;
+UPDATE `test`.`t1` SET `padding`='XXX' WHERE `k1`=1 AND `k2`=1;
+COMMIT;
+START TRANSACTION;
+UPDATE `test`.`t1` SET `padding`='YYY' WHERE `k1`=2 AND `k2`=2;
+UPDATE `test`.`t1` SET `padding`='YYY' WHERE `k1`=2 AND `k2`=3;
+COMMIT;
+START TRANSACTION;
+UPDATE `test`.`t1` SET `padding`='ZZZ' WHERE `k1`=2 AND `k2`=3;
+UPDATE `test`.`t1` SET `padding`='ZZZ' WHERE `k1`=3 AND `k2`=3;
+COMMIT;
+START TRANSACTION;
+UPDATE `test`.`t1` SET `padding`='AAA' WHERE `k1`=1 AND `k2`=1;
+UPDATE `test`.`t1` SET `padding`='AAA' WHERE `k1`=2 AND `k2`=2;
+UPDATE `test`.`t1` SET `padding`='AAA' WHERE `k1`=2 AND `k2`=3;
+UPDATE `test`.`t1` SET `padding`='AAA' WHERE `k1`=3 AND `k2`=3;
+COMMIT;
+START TRANSACTION;
+DROP TABLE `t1`;
+COMMIT;
+SET GLOBAL transaction_log_truncate_debug= true;

=== added file 'plugin/transaction_log/tests/t/create_schema-master.opt'
--- plugin/transaction_log/tests/t/create_schema-master.opt	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/t/create_schema-master.opt	2010-02-25 20:20:08 +0000
@@ -0,0 +1,1 @@
+--default-replicator-enable --transaction-log-enable --scheduler=multi_thread

=== added file 'plugin/transaction_log/tests/t/create_schema.inc'
--- plugin/transaction_log/tests/t/create_schema.inc	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/t/create_schema.inc	2010-02-25 20:20:08 +0000
@@ -0,0 +1,11 @@
+# 
+# Test CREATE SCHEMA and replication
+#
+
+--disable_warnings
+DROP SCHEMA IF EXISTS my_new_schema;
+--enable_warnings
+
+CREATE SCHEMA my_new_warnings;
+
+DROP SCHEMA my_new_warnings;

=== added file 'plugin/transaction_log/tests/t/create_schema.test'
--- plugin/transaction_log/tests/t/create_schema.test	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/t/create_schema.test	2010-02-25 20:20:08 +0000
@@ -0,0 +1,9 @@
+# Populate log with some records...
+--source ../plugin/transaction_log/tests/t/create_schema.inc
+
+# Read in the transaction.log.
+
+--exec ../drizzled/message/transaction_reader var/master-data/transaction.log
+
+# Truncate the log file to reset for the next test
+--source ../plugin/transaction_log/tests/t/truncate_log.inc

=== added file 'plugin/transaction_log/tests/t/multi_column_primary_key-master.opt'
--- plugin/transaction_log/tests/t/multi_column_primary_key-master.opt	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/t/multi_column_primary_key-master.opt	2010-02-25 20:20:08 +0000
@@ -0,0 +1,1 @@
+--default-replicator-enable --transaction-log-enable --scheduler=multi_thread

=== added file 'plugin/transaction_log/tests/t/multi_column_primary_key.inc'
--- plugin/transaction_log/tests/t/multi_column_primary_key.inc	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/t/multi_column_primary_key.inc	2010-02-25 20:20:08 +0000
@@ -0,0 +1,36 @@
+# 
+# Test that tables with multi-column primary keys
+# are handled properly in replication
+#
+
+--disable_warnings
+DROP TABLE IF EXISTS t1;
+--enable_warnings
+
+CREATE TABLE t1 (
+  k1 INT NOT NULL
+, k2 INT NOT NULL
+, padding VARCHAR(200) NOT NULL
+, PRIMARY KEY (k1, k2)
+);
+
+INSERT INTO t1 VALUES (1, 1, "I love testing.");
+INSERT INTO t1 VALUES (2, 2, "I hate testing.");
+INSERT INTO t1 VALUES (2, 3, "I hate and love testing.");
+INSERT INTO t1 VALUES (3, 3, "I adore testing.");
+
+# Simple PK update on both columns
+UPDATE t1 SET padding= "XXX" WHERE k1= 1 AND k2= 1;
+
+# UPDATE all records in table matching first column
+# in primary key
+UPDATE t1 SET padding= "YYY" WHERE k1= 2;
+
+# UPDATE all records in table matching second column
+# in primary key
+UPDATE t1 SET padding= "ZZZ" WHERE k2= 3;
+
+# UPDATE all records in table
+UPDATE t1 SET padding= "AAA";
+
+DROP TABLE t1;

=== added file 'plugin/transaction_log/tests/t/multi_column_primary_key.test'
--- plugin/transaction_log/tests/t/multi_column_primary_key.test	1970-01-01 00:00:00 +0000
+++ plugin/transaction_log/tests/t/multi_column_primary_key.test	2010-02-25 20:20:08 +0000
@@ -0,0 +1,9 @@
+# Populate log with some records...
+--source ../plugin/transaction_log/tests/t/multi_column_primary_key.inc
+
+# Read in the transaction.log.
+
+--exec ../drizzled/message/transaction_reader var/master-data/transaction.log
+
+# Truncate the log file to reset for the next test
+--source ../plugin/transaction_log/tests/t/truncate_log.inc

